from .dataset_styles import site_default_style


class ChartDataset:
    def __init__(self, values=None, style=None):
        if values is None:
            values = []

        if style is None:
            style = site_default_style

        self.values = values
        self.style = style

    def get_format(self, include_zero):
        values = [0] + self.values if include_zero else self.values
        return {**{"data": values}, **self.style.get_dictionary()}


class ScatterDataset(ChartDataset):
    def get_format(self, include_zero):
        values = [{"x": 0, "y": 0}] + self.values if include_zero else self.values
        return {**{"data": values}, **self.style.get_dictionary()}



def python_to_javascript_name(name: str):
    """
    Converts a Python variable name to a JavaScript name
    e.g. background_color -> backgroundColor
    """
    splitted_name = name.split("_")
    capitalized_name = [splitted_name[0]] + [name.title() for name in splitted_name[1:]]
    return "".join(capitalized_name)


class DatasetStyle:
    def __init__(self, **kwargs):
        self.styles = {}
        for style_name, value in kwargs.items():
            style_name_formatted = python_to_javascript_name(style_name)
            self.styles[style_name_formatted] = value

    def get_dictionary(self):
        return self.styles


default_styles = {
    "border_color": "rgba(255, 127, 80, 0.5)",
    "border_width": 2,
    "hover_background_color": "rgb(255, 127, 80)",
    "hover_border_width": 7,
    "background_color": [
        "rgba(255, 127, 80, 0.2)",
        "rgba(255, 127, 80, 0.3)",
        "rgba(255, 127, 80, 0.4)",
        "rgba(255, 127, 80, 0.5)",
        "rgba(255, 127, 80, 0.6)",
    ],
    "point_background_color": "orange",
    "point_border_width": 8,
}

site_default_style = DatasetStyle(**default_styles)

"""
Some common styles:

background_color,
border_cap_style,
border_color,
border_width,
fill,
hover_background_color,
hover_border_cap_style,
hover_border_color,
hover_border_width,
label,
line_tension,
point_background_color,
point_border_color,
point_border_width,
point_style,
"""



from django.middleware.common import MiddlewareMixin
from ..models import BrowserSession
import uuid


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def parse_user_agent(agent):
    browsers = {
        "edg": "Microsoft Edge",
        "chrome": "Chrome",
        "firefox": "Firefox",
        "op": "Opera",
        "opera": "Opera",
        "safari": "Safari",
        "instagram": "Instagram",
        "fban": "Facebook",
        "fbav": "Facebook",
        "msie": "Internet Explorer",
        "trident": "Internet Explorer",
        "snapchat": "Snapchat"
    }

    devices = {
        "iphone": "iPhone",
        "ipad": "iPad",
        "android": "Android",
        "windows": "Windows",
        "linux": "Linux",
        "macintosh": "Macintosh",
        "tesla": "Tesla"
    }

    agent = agent.lower()

    agent_info = {
        "browser": "",
        "device": "",
    }

    for browser in browsers.keys():
        if browser in agent:
            agent_info["browser"] = browsers[browser]
            break

    for device in devices.keys():
        if device in agent:
            agent_info["device"] = devices[device]
            break

    return agent_info


class StatMiddleware(MiddlewareMixin):
    def process_request(self, request):
        # get stat token cookie
        token = request.COOKIES.get("statsid")

        # if token is available, try to find existing BrowserSession instance by uuid
        if token:
            try:
                session = BrowserSession.objects.get(token=token)
            except BrowserSession.DoesNotExist:
                token = None
            else:
                request.stat_session = session

        # create new session instance if session doesn't exist
        if not token:
            new_token = uuid.uuid4()
            ip = get_client_ip(request)
            agent = request.META.get("User-Agent", "") if "User-Agent" in request.META.keys() else request.META.get("HTTP_USER_AGENT", "")

            agent_info = parse_user_agent(agent)

            session = BrowserSession.objects.create(
                token=new_token,
                ip=ip,
                agent=agent,
                browser=agent_info["browser"],
                device=agent_info["device"],
            )
            request.stat_session = session

    def process_response(self, request, response):
        response.set_cookie("statsid", request.stat_session.token)
        return response



from .types import *


class ChartManager:
    """
    Stores all of the charts for this website
    """
    def __init__(self):
        self.charts = []

    def add_chart(self, chart: Chart):
        """append chart to list"""
        self.charts.append(chart)


statistic_charts = ChartManager()



from django.db.models import *


class BrowserSession(Model):
    token = UUIDField(primary_key=True, unique=True)
    ip = GenericIPAddressField()
    agent = TextField()
    browser = CharField(max_length=50, null=True)
    device = CharField(max_length=50, null=True)
    start_time = DateTimeField(auto_now_add=True)


class PageSession(Model):
    page = TextField()
    browser_session = ForeignKey(BrowserSession, on_delete=CASCADE)
    new_to_website = BooleanField(default=None, null=True)
    referer = TextField(null=True)
    request_time = DateTimeField(auto_now_add=True)
    last_viewed = BooleanField(default=True)


class PageEvent(Model):
    page_session = ForeignKey(PageSession, on_delete=CASCADE)
    type = CharField(max_length=50)
    info = CharField(max_length=50)



from django.http import HttpResponse, HttpRequest, JsonResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from stats_dashboard.models import BrowserSession, PageSession, PageEvent
from stats_dashboard.charts.manager import statistic_charts
from django.conf import settings
import importlib
import json
import re


def stats_dashboard(request):
    if not request.user.is_staff:
        return HttpResponse(status=401)
    return render(request, "stats/index.html", {
        "browser_sessions": BrowserSession.objects,
        "page_session": PageSession.objects,
        "page_events": PageEvent.objects,
    })


@csrf_exempt
def activity(request: HttpRequest):
    """
    Any user activity is sent as a request here
    """
    if request.method != "POST":
        # only post requests are allowed
        # HTTP 405 is an invalid method
        response = HttpResponse(status=405)
        # set allow header to POST
        response["Allow"] = "POST"
        return response

    if not settings.get("STAT_TRACK_IN_DEBUG", default=True) and settings.DEBUG:
        return HttpResponse()

    data = json.loads(request.body)

    # get all pages with a foreign key to the current browser session
    # stat session is added at middleware
    pages = request.stat_session.pagesession_set

    if data["type"] == "navigate":
        referer = request.META.get("Referer", None)
        # on a NEW navigation, create page session instance
        to = data["info"]

        create_first = False
        # get last viewed page session
        try:
            last_viewed = pages.get(last_viewed=True)
        except PageSession.DoesNotExist:
            create_first = True

        # check whether the new page is different to previous, else could be refresh
        if create_first or last_viewed.page != to:
            if not create_first:
                last_viewed.last_viewed = False
                last_viewed.save()
            new_page = PageSession.objects.create(
                page=to,
                browser_session=request.stat_session,
                referer=referer,
                last_viewed=True
            )

    elif data["type"] == "new_to_website":
        try:
            last_viewed = pages.filter(last_viewed=True)
            # once new to website is true, can't change back. stops against refresh overriding a true value to false
            if not last_viewed[0].new_to_website:
                last_viewed.update(new_to_website=data["info"])
        except:
            pass

    elif data["type"] == "click":
        try:
            PageEvent.objects.create(
                page_session=pages.get(last_viewed=True),
                type="click",
                info=data["info"],
            )
        except PageSession.DoesNotExist:
            pass

    return HttpResponse()


def config(request):
    """
    Returns the config for a given path, e.g. what to send to activity api
    """
    if request.method != "GET":
        # only get requests are allowed
        # HTTP 405 is an invalid method
        response = HttpResponse(status=405)
        # set allow header to GET
        response["Allow"] = "GET"
        return response

    path = request.GET["page"]

    config_for_path = {
        "click": [],
    }

    try:
        settings.STATS_PAGE_CONFIG
    except AttributeError:
        return JsonResponse(config_for_path)

    for re_path, configs in settings.STATS_PAGE_CONFIG.items():
        if re_path == "*" or re.match(re_path, path):
            for config_name, config_value in configs.items():
                config_for_path[config_name] = config_value

    return JsonResponse(config_for_path)


def all_charts(request: HttpRequest):
    if not request.user.is_staff:
        return JsonResponse([], safe=True)

    try:
        chart_file = importlib.import_module(settings.STATS_CHART_LOCATION)
    except ModuleNotFoundError:
        return JsonResponse([], safe=True)

    charts = statistic_charts.charts
    json_charts = []
    for chart in charts:
        json_charts.append(chart.get_json_data())

    return JsonResponse(json_charts, safe=False)



from setuptools import setup, find_packages

desc = open("README.md", "r")
desc_text = desc.read()
desc.close()

setup(
    name="django-statistics-dashboard",
    version="0.4",
    description="A stats dashboard and session tracker for Django.",
    long_description=desc_text,
    long_description_content_type="text/markdown",
    url="https://github.com/Si1veR123/django-statistics-dashboard",
    author="Connor George",
    author_email="connorgeorgeem@aol.com",
    license="MIT",
    packages=find_packages(),
    include_package_data=True
)



from .types import LineChart, PieChart, BarChart
from .dataset import ChartDataset
from stats_dashboard.models import BrowserSession, PageSession
import datetime, calendar


class InvalidChartType(Exception):
    pass


def visits_time(visit_time=15768000, scale="month", style=None) -> LineChart:
    """
    load the 'visits over time' line chart, which displays
    browser sessions over a set period of time to a set scale
    """

    allowed_scales = ("year", "month", "week", "day", "hour", "minute")

    if scale not in allowed_scales[:-1]:
        raise Exception("Invalid time scale. Should be in ", str(allowed_scales[:-1]))

    if visit_time != 0:
        # get the start time for the graph from the chart visit time
        first_time = datetime.datetime.today() - datetime.timedelta(seconds=visit_time)

        # set the time to the start of its scale e.g. if scale is month, reset day to 1
        if scale == "month":
            time_replacement_kwargs = {"day": 1}
        else:
            time_replacement_kwargs = {allowed_scales[allowed_scales.index(scale) + 1]: 1}

        first_time = first_time.replace(tzinfo=datetime.timezone.utc, **time_replacement_kwargs)

    else:
        # if it is 0, get all visits
        first_time = datetime.datetime.fromtimestamp(0)

    # filter the browser session to greater than the start time
    visits_time_queryset = BrowserSession.objects.filter(start_time__gte=first_time)
    # get each record's start time
    x_unfiltered = [time["start_time"] for time in visits_time_queryset.values("start_time")]

    # y and x axis values, populated later
    y = []
    x = []

    # current time is used to create the x axis scale. starts at the first time
    current_time = first_time

    # if seeing all visits, set the current time to the earliest browser session
    if visit_time == 0:
        earliest_visit = datetime.datetime.now(datetime.timezone.utc)
        for visit in x_unfiltered:
            if visit < earliest_visit:
                earliest_visit = visit
        current_time = earliest_visit

    # this loop adds times to the x axis values from start time to now, in intervals of the set scale
    while datetime.datetime.now(datetime.timezone.utc) >= current_time:
        x.append(current_time)
        if scale == "month":
            # add number of days in current month
            current_time += datetime.timedelta(days=calendar.monthrange(current_time.year, current_time.month)[1])
        elif scale == "year":
            # add number of days in current year
            current_time += datetime.timedelta(days=366 if calendar.isleap(current_time.year) else 365)
        else:
            # add 1 of the set scale, e.g. if scale is week, kwargs is weeks=1
            time_add_kwargs = {scale + "s": 1}
            current_time += datetime.timedelta(**time_add_kwargs)
    x.append(current_time)

    # for each time interval check each browser session and create a count of sessions in that interval
    for label in x:
        count = 0
        for date in x_unfiltered:
            if label.year == date.year:
                if scale == "year":
                    count += 1
                    continue

                # check week without month as weeks can overlap months
                if label.isocalendar()[1] == date.isocalendar()[1]:
                    if scale == "week":
                        count += 1
                        continue

                if label.month == date.month:
                    if scale == "month":
                        count += 1
                        continue
                    if label.day == date.day:
                        if scale == "day":
                            count += 1
                            continue
                        if label.hour == date.hour:
                            if scale == "hour":
                                count += 1
                                continue

        y.append(count)

    visits_time_chart = LineChart(
        name="Visits over time" + " ({})".format(scale),
        labels=x,
        datasets=[ChartDataset(y, style)],
        time_unit=scale
    )

    return visits_time_chart


def browser_share(style=None, type="pie"):

    browser_count = {"Other/Unknown": 0}

    browsers = [b["browser"] for b in BrowserSession.objects.all().values("browser")]

    for browser in browsers:
        if browser is None or browser == "":
            browser_count["Other/Unknown"] += 1
            continue

        if browser not in browser_count.keys():
            browser_count[browser] = 0
        browser_count[browser] += 1

    x = list(browser_count.keys())
    y = list(browser_count.values())

    if type == "pie":
        ChosenChart = PieChart
    elif type == "bar":
        ChosenChart = BarChart
    else:
        raise InvalidChartType

    browser_share_chart = ChosenChart(
        name="Browser Share",
        labels=x,
        datasets=[
            ChartDataset(y, style),
        ]
    )

    return browser_share_chart


def device_share(style=None, type="pie"):

    device_count = {"Other/Unknown": 0}

    devices = [b["device"] for b in BrowserSession.objects.all().values("device")]

    for device in devices:
        if device is None or device == "":
            device_count["Other/Unknown"] += 1
            continue

        if device not in device_count.keys():
            device_count[device] = 0
        device_count[device] += 1

    x = list(device_count.keys())
    y = list(device_count.values())

    if type == "pie":
        ChosenChart = PieChart
    elif type == "bar":
        ChosenChart = BarChart
    else:
        raise InvalidChartType

    device_share_chart = ChosenChart(
        name="Device Share",
        labels=x,
        datasets=[
            ChartDataset(y, style),
        ]
    )

    return device_share_chart


def common_pages(style=None):
    all_pages = PageSession.objects.all()

    all_urls = [page["page"] for page in all_pages.values("page")]

    page_count = {}

    for url in all_urls:
        if url not in page_count.keys():
            page_count[url] = 0
        page_count[url] += 1

    common_pages_chart = BarChart(
        name="Common Pages",
        labels=list(page_count.keys()),
        datasets=[
            ChartDataset(list(page_count.values()), style)
        ],
        options={
            "scales": {
                "yAxes": [{
                    "ticks": {
                        "stepSize": 1
                    }
                }]
            }
        }
    )

    return common_pages_chart



import typing
from .dataset import ChartDataset


class Chart:
    def __init__(
                self,
                name: str = "",
                labels: list = None,
                datasets: typing.List[ChartDataset] = None,
                options: dict = None,
                include_zero: bool = False
            ):

        self.name = name
        self.type = ""
        self.include_zero = include_zero

        if labels is None:
            labels = []
        self.labels = labels

        if include_zero:
            self.labels.insert(0, 0)

        if datasets is None:
            datasets = []
        self.datasets = datasets

        if not options:
            self.options = {}
        else:
            self.options = options

        legend_options = self.options.get("legend")
        if legend_options:
            try:
                legend_options["display"] = False
            except:
                pass
        else:
            self.options["legend"] = {"display": False}

    def get_json_data(self):
        json_data = {
            "data": {
                "labels": self.labels,
                "datasets": [dataset.get_format(self.include_zero) for dataset in self.datasets]
            },
            "options": self.options,
            "type": self.type,
            "name": self.name
        }

        return json_data


class BarChart(Chart):
    def __init__(
                self,
                name: str = "",
                labels: list = None,
                datasets: typing.List[ChartDataset] = None,
                options: dict = None,
                include_zero: bool = False
            ):

        super().__init__(name, labels, datasets, options, include_zero)
        self.type = "bar"


class LineChart(Chart):
    def __init__(
                self,
                name: str = "",
                labels: list = None,
                datasets: typing.List[ChartDataset] = None,
                options: dict = None,
                include_zero: bool = False,
                time_unit: str = None
            ):

        super().__init__(name, labels, datasets, options, include_zero)
        self.type = "line"
        self.time_unit = time_unit

        if time_unit is None:
            return

        try:
            self.options["scales"]
        except KeyError:
            self.options["scales"] = {}

        try:
            self.options["scales"]["xAxes"]
        except KeyError:
            self.options["scales"]["xAxes"] = []

        self.options["scales"]["xAxes"] = self.options["scales"]["xAxes"] + [{"type": "time", "time": {"unit": self.time_unit}}]


class PieChart(Chart):
    def __init__(
                self,
                name: str = "",
                labels: list = None,
                datasets: typing.List[ChartDataset] = None,
                options: dict = None,
                include_zero: bool = False,
            ):

        super().__init__(name, labels, datasets, options, include_zero)
        self.type = "pie"


class ScatterChart(Chart):
    def __init__(
                self,
                name: str = "",
                labels: list = None,
                datasets: typing.List[ChartDataset] = None,
                options: dict = None,
                include_zero: bool = False,
            ):

        super().__init__(name, labels, datasets, options, include_zero)
        self.type = "scatter"

